#include "LandmarkDetectionGenerator.h"
#include "lawrence/local_interface.h"
#include <list>

#define LMDG_DEBUG

//better to put this here instead of the header
using namespace std; 

typedef struct sqInfo {
	const VisionObjectEvent* square;
	const LMColors clr;
	
	sqInfo(const VisionObjectEvent* obj, const LMColors color):
		square(obj),
		clr(color) {}
} sqInfo;

typedef list<LMColors> COLORLST;
typedef list<sqInfo> SQLST;

LandmarkDetectionGenerator::LandmarkDetectionGenerator() :
	BehaviorBase("LandmarkDetectionGenerator"),
	confidence(0),
	node(NULL) {
	for(unsigned int i=0; i<NUM_SQUARES; i++) squares[i] = NULL;
	}

void LandmarkDetectionGenerator::DoStart() {
	BehaviorBase::DoStart(); // do this first (required)
	erouter->addListener(this,EventBase::visObjEGID);
}

void LandmarkDetectionGenerator::DoStop() {
	BehaviorBase::DoStop(); // do this last (required)
}

void LandmarkDetectionGenerator::processEvent(const EventBase& e) {
	// First ignor self generated events.
	if(e.getSourceID() == LawrenceInterface::visLMSID) return;
	if(e.getGeneratorID() == EventBase::visObjEGID) {
		VisionObjectEvent* visE = new VisionObjectEvent(dynamic_cast<const VisionObjectEvent&>(e));
		static int ind;
		switch(e.getSourceID()) {
		case LawrenceInterface::visLMPSID:
			ind = PINK;
			break;
		case LawrenceInterface::visLMBSID:
			ind = BLUE;
			break;
		case LawrenceInterface::visLMYSID:
			ind = YELLOW;
			break;
		case LawrenceInterface::visLMGSID:
			ind = GREEN;
			break;
		default:
			ind = -1;
		}
		if(ind > -1) {
			if(squares[ind] != NULL) delete squares[ind];
			squares[ind] = visE;
		}

		if(node != NULL) delete node;
		node = determineLandmark();
		// If it does look like a landmark
		if(node != NULL) { 
			erouter->postEvent(EventBase::visObjEGID,LawrenceInterface::visLMSID,EventBase::activateETID,0);
#ifdef LMDG_DEBUG
			cout << "Landmark detected:" << endl;
			node->print();
#endif
		}
	}	
	else {
		cout << "LandmarkDetectionGenerator got unexpected event: "
			 << e.getName() << endl;
	}
}

MapNode* LandmarkDetectionGenerator::getNode() { return node; }

LMColors LandmarkDetectionGenerator::colorByElimination(LMColors a, LMColors b, LMColors c){
	if((a != PINK) && (b != PINK) && (c != PINK)) return PINK;
	else if((a != BLUE) && (b != BLUE) && (c != BLUE)) return BLUE;
	else if((a != YELLOW) && (b != YELLOW) && (c != YELLOW)) return YELLOW;
	else return GREEN;
}

/*********************************************
 * BEGIN UGLY LANDMARK RECOGNITION CODE!!!!! *
 ********************************************/


// This is where it get's complicated, very
MapNode* LandmarkDetectionGenerator::determineLandmark() {	
	// Start with (relitively) quick checks to throw stuff out
	
	// If there's no pink square we won't even try since we base
	// several calculations off the pink square
	if(squares[PINK] == NULL) return NULL;
	/* Area calculations are based off of pink because pink is the most
	 * reliable color, we're almost garunteed that pink we see is from
	 * an actual landmark */
	float refArea = squares[PINK]->getBoundaryArea();
	unsigned int pulse = squares[PINK]->getFrame();
	unsigned short numSqSeen = 0;
	SQLST nonNullSqs;
	for(unsigned int i=0; i<NUM_SQUARES; i++) {
		if(squares[i] != NULL) {
			// Squares must be in the same frame
			if(squares[i]->getFrame() != pulse) {
#ifdef LMDG_DEBUG
				cout << "LMDG: failed frame offset limit" << endl;
#endif
				delete squares[i]; // Remove the old square
				squares[i] = NULL;
				return NULL;
			}
			numSqSeen++; // Saw another square
			nonNullSqs.push_front(sqInfo(squares[i],(LMColors)i));
		}
	}
	// There must be at least 3 squares
	if(numSqSeen < 3) {
#ifdef LMDG_DEBUG
		cout << "LMDG: failed number of squares seen test" << endl;
#endif
		return NULL; 
	}
	// End quick checks, begin hard part

	confidence = 1; // Start at 1 and go down
	// Will lists of the colors from ordered from highest to left to right;
	COLORLST up, left;
	// Will have upper-left, upper-right, ... , colors once we've figured them out
	LMColors ul, ur, ll, lr; 
	// Iterators we'll be using over and over
	COLORLST::iterator itr; 
	SQLST::iterator sq;
	
	for(sq=nonNullSqs.begin(); sq!=nonNullSqs.end(); sq++) {
		/* Reduce confidence by difference in area
		 *      | thisArea - refArea |
		 * 1 - ------------------------
		 *             refArea
		 */
//		confidence *= 1.0f - (abs(sq->square->getBoundaryArea() - refArea) / refArea);

		// Add to list ordered by Y position
		for(itr=up.begin(); itr!=up.end(); itr++) {
			// If this square is higher than the itr square
			if(sq->square->getCenterY() < squares[*itr]->getCenterY()) {
				up.insert(itr, sq->clr);
				break; // No need to loop further
			}
		}
		// If it wasn't added in the loop add it to the end
		if(itr == up.end()) up.push_back(sq->clr);
		// Add to the list ordered by X position
		for(itr=left.begin(); itr!=left.end(); itr++) {
			if(sq->square->getCenterX() < squares[*itr]->getCenterX()) {
				up.insert(itr, sq->clr);
				break; // No need to loop further
			}
		}
		// If it wasn't added in the loop add it to the end
		if(itr == left.end()) left.push_back(sq->clr);
	}

#ifdef LMDG_DEBUG
	cout << "Y ordered list:\t";
	for(itr=up.begin(); itr!=up.end(); itr++) cout << *itr << "\t";
	cout << endl << "X ordered list:\t";
	for(itr=left.begin(); itr!=left.end(); itr++) cout << *itr << "\t";
	cout << endl;
#endif

	// Saw 3 squares, extrapolate the 4th case
	if(numSqSeen == 3) {
		cout << "LMDG: Extrapolating 4th square" << endl;
		// Reduce confidence in accordance with the number of squares missing
		//confidence *= 0.75d;
		
		float distAB, distBC, distCA;
		distAB = (nonNullSqs.front())->square->getCenterX() - 
		
		return NULL; // For test just return null
	}
	// Saw full 4 squares case
	else {
		// Figure out the upper left color
		// Upper most and left most case
		if(up.front() == left.front()) ul = up.front();
		// Upper most and second left most
		else if(up.front() == *(++(left.begin()))) ul = up.front();
		// Second upper most and left most
		else ul = left.front();
		
		// Figure out upper right color
		// Upper most and right most case
		if(up.front() == left.back()) ur = up.front();
		// Upper most and second right most case
		else if(up.front() == *(--(left.begin()))) ul = up.front();
		// Second upper most and right most
		else ul = left.back();
		
		// Figure out lower left color
		// Lowest and left most case
		if(up.back() == left.front()) ll = left.front();
		// Lowest and second left most
		else if(up.back() == *(++(left.begin()))) ll = up.back();
		// Second lowest and left most
		else ll = left.front();
		
		// Figure out the lr color
		// This is the only color left
		lr = colorByElimination(ul, ur, ll);
	}
	return new MapNode(ul, ur, ll, lr);
}
