#include "SquareDetectionGenerator.h"
#include "Events/EventRouter.h"
#include "Shared/Util.h"
#include "Shared/WorldState.h"
#include "Wireless/Wireless.h"
#include "Shared/Config.h"
#include "Events/VisionObjectEvent.h"
#include "Vision/SegmentedColorGenerator.h"
#include "Vision/RegionGenerator.h"

typedef RegionGenerator::region region; //!< shorthand - using CMVision regions
typedef SegmentedColorGenerator::color_class_state color_class_state; //!< shorthand - using CMVision color structs

SquareDetectionGenerator::SquareDetectionGenerator(unsigned int mysid, const RegionGenerator * rg, unsigned int colorIdx, unsigned int threshmapChan, unsigned int noiseFiltering, float confidence)
	: EventGeneratorBase("SquareDetectionGenerator","SquareDetectionGenerator",EventBase::visObjEGID,mysid,rg->getGeneratorID(),rg->getSourceID(),EventBase::statusETID), clrIdx(colorIdx), tmIdx(threshmapChan), square(), present(false), count(0), noiseThreshold(noiseFiltering), confidenceThreshold(confidence)
{}

void
SquareDetectionGenerator::processEvent(const EventBase& e) {
	PROFSECTION("SquareDetection::processEvent()",state->mainProfile);
	EventGeneratorBase::processEvent(e);
	if(e.getGeneratorID()!=getListenGeneratorID() || e.getSourceID()!=getListenSourceID())
		return;

	const SegmentedColorFilterBankEvent * segev=dynamic_cast<const SegmentedColorFilterBankEvent*>(&e);
	if(NULL==segev) {
		serr->printf("SquareDetectionGenerator's event %s was not a SegmentedColorFilterBankEvent",e.getName().c_str());
		return;
	}
	
  static const bool debug_square = false;
  static const bool debug_conf = false;

  static int frame_cnt=0;
  static const int print_period=1;

  if(debug_square)
    frame_cnt = (frame_cnt + 1) % print_period;

	unsigned int layer=segev->getNumLayers()-config->vision.resolution-1;
	if(segev->getNumColors()<=clrIdx) {
		serr->printf("SquareDetectionGenerator::clrIdx %d is invalid (only %d available)\n",clrIdx,segev->getNumColors());
		return;
	}
	if(segev->getNumChannels()<=tmIdx) {
		serr->printf("SquareDetectionGenerator::tmIdx %d is invalid (only %d available)\n",tmIdx,segev->getNumChannels());
		return;
	}
	if(segev->getNumLayers()<=layer) {
		serr->printf("SquareDetectionGenerator's selected layer %d is invalid (only %d available)\n",layer,segev->getNumLayers());
		return;
	}
	if(segev->getImage(layer,tmIdx)==NULL) {
		serr->printf("SquareDetectionGenerator source's getImage returned NULL!\n");
		return;
	}
  const color_class_state& squareCCS=reinterpret_cast<const color_class_state*>(segev->getImage(layer,tmIdx))[clrIdx];

  square.confidence = 0;
  region * square_region=NULL;

  region * or_reg=squareCCS.list; //head of linked list of regions of target color. or_reg= original region (without noise removal)
  if(!or_reg){
		count++;
		// The next 3 lines could be cleared if no dimensions are sent with the deactivation events
		float dim=segev->getWidth(layer)>segev->getHeight(layer) ? segev->getWidth(layer) : segev->getHeight(layer);	
		//the max width and height of scaled dimensions
		float cw=segev->getWidth(layer)/dim;
		float ch=segev->getHeight(layer)/dim;	
		unsigned int frame_number=segev->getFrameNumber();
		if (present && count>noiseThreshold) {  // If there are no regions a number of frames and square was present
			count=0;
			present=false; // square is not present
			createEvent(EventBase::deactivateETID,0,0,0,0,0,cw,ch,frame_number); // Deactivation Events
		}
		return;
  }
  
  unsigned int n = 0;
  while(or_reg && n<NUM_CHECK) {
    //float conf,conf0,conf_square_bbox,conf_area,conf_green,conf_area_bonus;
    //float conf_red_v_area;
    //int edge;

    int w = or_reg->x2 - or_reg->x1 + 1;
    int h = or_reg->y2 - or_reg->y1 + 1;
    
    int edge = calcEdgeMask(or_reg->x1,or_reg->x2,or_reg->y1,or_reg->y2,segev->getWidth(layer),segev->getHeight(layer));
    float conf0 = (w >= 3) * (h >= 3) * (or_reg->area >= 7);
    float conf_square_bbox = 
      edge ?
      gaussian_with_min(pct_from_mean(w,h) / .6f, 1e-3) :
      gaussian_with_min(pct_from_mean(w,h) / .2f, 1e-3);
    float conf_area =
      edge ?
      gaussian_with_min(pct_from_mean(((float)M_PI)*w*h/4.0f,or_reg->area) / .6f, 1e-3) :
      gaussian_with_min(pct_from_mean(((float)M_PI)*w*h/4.0f,or_reg->area) / .2f, 1e-3);
    float conf_area_bonus = or_reg->area / 1000.0f;

    float conf = conf0*conf_square_bbox*conf_area + conf_area_bonus;

    if(conf > 1.0) conf = 1.0f;
    
    if(debug_conf && frame_cnt == 0) {
      printf("conf0 %g conf_square_bbox %g conf_area %g conf_area_bonus %g final %g\n",
						 conf0,conf_square_bbox,conf_area,conf_area_bonus,conf);
    }
    
    if(conf > square.confidence) {
		square.confidence = conf;
		square_region = or_reg;
    }

    or_reg = or_reg->next;
    n++;
  }  
	
	testSendEvent(*segev,square.confidence,(int)square_region->x1,(int)square_region->x2,(int)square_region->y1,(int)square_region->y2,(int)square_region->area); 
}

void
SquareDetectionGenerator::testSendEvent(const FilterBankEvent& ev, float conf, int regX1,int regX2,int regY1,int regY2, int area) {
	unsigned int layer=ev.getNumLayers()-config->vision.resolution-1;

	//pick the larger dimension for scaling, one dimension will be [-1,1], the other will depend on aspect ratio
	float dim=ev.getWidth(layer)>ev.getHeight(layer) ? ev.getWidth(layer) : ev.getHeight(layer);

	//the max width and height of scaled dimensions
	float cw=config->vision.x_range; //ev.getWidth(layer)/dim;
	float ch=config->vision.y_range; //ev.getHeight(layer)/dim;

	//scale the dimensions
	float cx1=2.0f*regX1/dim - cw;
	float cx2=2.0f*(regX2+1)/dim - cw;
	float cy1=2.0f*regY1/dim - ch;
	float cy2=2.0f*(regY2+1)/dim - ch;
	unsigned int frame_number=ev.getFrameNumber();	

	if (conf>confidenceThreshold) {
		if (present) {
			count=0;
			createEvent(EventBase::statusETID,cx1,cx2,cy1,cy2,area,cw,ch,frame_number);
		} else {
			count++;
			if (count>noiseThreshold) {
				count=0;
				present=true;
				createEvent(EventBase::activateETID,cx1,cx2,cy1,cy2,area,cw,ch,frame_number);
			}
		}
	} else {
		if (!present) {
			count=0;
		} else {
			count++;
			if (count>noiseThreshold) {
				count=0;
				present=false;
				createEvent(EventBase::deactivateETID,0,0,0,0,area,cw,ch,frame_number);
			} else {
				createEvent(EventBase::statusETID,cx1,cx2,cy1,cy2,area,cw,ch,frame_number);
			}
		}
	}
}

void
SquareDetectionGenerator::createEvent(EventBase::EventTypeID_t etid, float bbX1,float bbX2,float bbY1,float bbY2,float area,float rx,float ry,unsigned int frame) const {
	VisionObjectEvent * vo=new VisionObjectEvent(mySourceID,etid,bbX1,bbX2,bbY1,bbY2,area,rx,ry,frame);
	vo->setName(getName());
	erouter->postEvent(vo);
}

int
SquareDetectionGenerator::calcEdgeMask(int x1,int x2,int y1,int y2, int width, int height) {
  static const int boundary_pixel_size=1;

  int edge = 0;
  if(x1 <= 0       +boundary_pixel_size) edge |= OFF_EDGE_LEFT  ;
  if(x2 >= width -1-boundary_pixel_size) edge |= OFF_EDGE_RIGHT ;
  if(y1 <= 0       +boundary_pixel_size) edge |= OFF_EDGE_TOP   ;
  if(y2 >= height-1-boundary_pixel_size) edge |= OFF_EDGE_BOTTOM;

  return edge;
}

/*! @file 
 * @brief Implements SquareDetectionGenerator, which uses segmented color region information to detect square objects
 * @author alokl (Creator)
 * @author ejt (reorganized)
 * @author Ignacio Herrero Reder < nhr at dte uma es > (VisionObjectInfo Boundary Box - bug 74)
 * @author Daniel Casner (Modified to detect squares)
 *
 * $Author: DTC $
 * $Name: tekkotsu-2_4_1 $
 * $Revision: 0.1 $
 * $State: Dvl $
 * $Date: 2006/04/12 18:24:15 $
 */

